"""
Helper methods to trigger tasks in asynchronous workers
"""
from typing import Optional, Union
from uuid import UUID

from arkindex.dataimport import tasks as dataimport_tasks
from arkindex.dataimport.models import DataImport, WorkerActivityState, WorkerVersion
from arkindex.documents import export
from arkindex.documents import tasks as documents_tasks
from arkindex.documents.managers import ElementQuerySet
from arkindex.documents.models import Corpus, CorpusExport, Element


def corpus_delete(corpus: Union[Corpus, UUID, str], user_id: Optional[int] = None) -> None:
    """
    Delete a whole corpus without killing a server by removing all related
    models first and lowering the amount of foreign keys Django has to handle.
    """
    if isinstance(corpus, Corpus):
        corpus_id = str(corpus.id)
        description = f'Deletion of corpus {corpus.name}'
    else:
        corpus_id = str(corpus)
        description = f'Deletion of corpus {corpus_id}'

    documents_tasks.corpus_delete.delay(corpus_id=corpus_id, description=description, user_id=user_id)


def element_trash(queryset: ElementQuerySet,
                  delete_children: bool = True,
                  user_id: Optional[int] = None) -> None:
    """
    Run ElementQuerySet.trash to delete a batch of elements.
    """
    assert isinstance(queryset, ElementQuerySet), 'Only Element querysets can be trashed'
    documents_tasks.element_trash.delay(
        queryset=queryset,
        delete_children=delete_children,
        user_id=user_id,
        description='Element deletion',
    )


def worker_results_delete(corpus_id: UUID,
                          version: Optional[WorkerVersion] = None,
                          parent_id: Optional[UUID] = None,
                          user_id: Optional[int] = None) -> None:
    """
    Delete all Worker Results produced by a specific WorkerVersion on a whole corpus or under
    a specified parent element (parent element included).
    """
    if version is None:
        description = 'Worker results deletion'
    else:
        description = f"Deletion of results produced by worker version: {version}"

    documents_tasks.worker_results_delete.delay(
        corpus_id=corpus_id,
        version_id=version.id if version else None,
        parent_id=parent_id,
        user_id=user_id,
        description=description,
    )


def move_element(source: Element, destination: Element, user_id: Optional[int] = None) -> None:
    """
    Move a source Element (and all of its children) to a destination Element.
    Delete all source Element previous parents.
    """
    documents_tasks.move_element.delay(
        source=source,
        destination=destination,
        user_id=user_id,
        description=f"Moving element {source.name} to element {destination.name}"
    )


def initialize_activity(process: DataImport):
    """
    Initialize activity on every process elements for worker versions that are part of its workflow
    """
    dataimport_tasks.initialize_activity.delay(process)


def export_corpus(corpus_export: CorpusExport) -> None:
    """
    Export a corpus to a SQLite database
    """
    export.export_corpus.delay(
        corpus_export=corpus_export,
        user_id=corpus_export.user_id,
        description=f'Export of corpus {corpus_export.corpus.name}'
    )


def process_delete(process: DataImport, delete_activity: Optional[WorkerActivityState] = None):
    """
    Delete a process with its related worker activities
    """
    dataimport_tasks.process_delete.delay(process, delete_activity=delete_activity)
